<?php
/*
 * imdr
 * @author Shinsuke Takeuchi <ctrans@gmail.com>
 * @license MIT License
 * @link http://www.karak.jp/
 * @version 1.01
 */
class imdr {
	private $goto;
	private $out;
	private $unum;

	public function __construct() {
		$this->goto = array();
		$this->out = array();
		$this->unum = 0;
	}

	public function add($word, $value) {
		$state = 0;
		$flg = FALSE;
		$enc = "UTF-8";
		$len = mb_strlen($word, $enc);
		#gotoが定義されているところまでstateをたどる
		for ($i=0; $i < $len; $i++) {
			$char = mb_substr($word, $i, 1, $enc);
			$key = $char . strval($state);
			if (array_key_exists($key, $this->goto)) {
				$state = $this->goto[$key];
			} else {
				$flg = TRUE;
				break;
			}
		}
		#フラグがFALSEである場合は、与えられた単語がこれまで定義された単語より
		#短いことを意味するため、新しいoutを定義して処理を終了する。
		if ($flg == FALSE) {
			$this->out[$state] = $value;
			return TRUE;
		}
		#単語の終端まで新たなgotoを定義し、最後にoutを定義する
		for ($j=$i; $j < $len; $j++) {
			$this->unum++;
			$char = mb_substr($word, $j, 1, $enc);
			$key = $char . strval($state);
			$this->goto[$key] = $this->unum;
			$state = $this->unum;
		}
		$this->out[$state] = $value;
		return TRUE;
	}

	#指定単語に対応するデータを辞書から削除
	public function remove($word) {
		#登録されていない単語の場合はFALSEを返す
		$state = $this->getstate($word);
		if ($state == -1) {return FALSE;}
		unset($this->out[$state]);
		return TRUE;
	}

	#キー（$state）に対応するoutの値を返す
	#Ver1.01 関数名変更getout->getvalue
	public function getvalue($state) {
		if (array_key_exists($state, $this->out)) {
			return $this->out[$state];
		}
		return;
	}

	#文字列に対応するstateを返す
	#存在しない場合は-1を返す
	public function getstate($str) {
		$state = 0;
		$enc = "UTF-8";
		$len = mb_strlen($str, $enc);
		for ($i=0; $i < $len; $i++) {
			$char = mb_substr($str, $i, 1, $enc);
			$key = $char . strval($state);
			if (array_key_exists($key, $this->goto)) {
				$state = $this->goto[$key];
			} else {
				return -1;
			}
		}
		if (!array_key_exists($state, $this->out)) {
			return -1;
		}
		return $state;
	}

	public function getlongest($target) {
		$state = 0;
		$i = 0;
		$match = -1;
		$pos = -1;
		$enc = "UTF-8";
		$len = mb_strlen($target, $enc);
		#gotoが定義されているところまでstateをたどる
		for ($i=0; $i < $len; $i++) {
			$char = mb_substr($target, $i, 1, $enc);
			$key = $char . strval($state);
			if (array_key_exists($key, $this->goto)) {
				$state = $this->goto[$key];
				if (array_key_exists($state, $this->out)) {
					$match = $state;
					$pos = $i;
				}
			} else {
				break;
			}
		}
		return array($match, $pos);
	}
	#targetに含まれる共通接頭辞を取得する
	#返値：共通接頭辞の配列
	public function commonprefix($target) {
		$state = 0;
		$matches = array();
		$enc = "UTF-8";
		$len = mb_strlen($target, $enc);
		for ($i=0; $i < $len; $i++) {
			$char = mb_substr($target, $i, 1, $enc);
			$key = $char . strval($state);
			if (array_key_exists($key, $this->goto)) {
				$state = $this->goto[$key];
				if (array_key_exists($state, $this->out)) {
					$matches[] = mb_substr($target, 0, $i+1, $enc);
				}
			} else {
				break;
			}
		}
		return $matches;
	}

	#Ver1.01 getprefix 関数追加
	#targetに含まれる共通接頭辞と対応するstateを取得する
	#返値：配列[単語]＝state
	public function getprefix($target) {
		$state = 0;
		$matches = array();
		$enc = "UTF-8";
		$len = mb_strlen($target, $enc);
		for ($i=0; $i < $len; $i++) {
			$char = mb_substr($target, $i, 1, $enc);
			$key = $char . strval($state);
			if (array_key_exists($key, $this->goto)) {
				$state = $this->goto[$key];
				if (array_key_exists($state, $this->out)) {
					$frag = mb_substr($target, 0, $i+1, $enc);
					$matches[$frag] = $state;
				}
			} else {
				break;
			}
		}
		return $matches;
	}

	#textに含まれる最長接頭辞を対応する値で置換する
	public function replace($text) {
		$result = array();
		$enc = "UTF-8";
		while (mb_strlen($text, $enc) > 0) {
			list($state, $pos) = $this->getlongest($text);
			if ($state != -1 && array_key_exists($state, $this->out)) {
				$result[] = $this->out[$state];
				$text = mb_substr($text, $pos+1, null, $enc);
			} else {
				$result[] = mb_substr($text, 0, 1, $enc);
				$text = mb_substr($text, 1, null, $enc);
			}
		}
		return $result;
	}
	#textに含まれる最長接頭辞と対応する値のペアを取得する
	public function getpair($text) {
		$result = array();
		$enc = "UTF-8";
		while (mb_strlen($text, $enc) > 0) {
			list($state, $pos) = $this->getlongest($text);
			if ($state != -1 && array_key_exists($state, $this->out)) {
				$origin = mb_substr($text, 0, $pos + 1, $enc);
				$replace = $this->getout($state);
				$result[] = array($origin, $replace);
				$text = mb_substr($text, $pos + 1, null, $enc);
			} else {
				$cur = mb_substr($text, 0, 1, $enc);
				$result[] = array($cur, $cur);
				$text = mb_substr($text, 1, null, $enc);
			}
		}
		return $result;
	}

	#textに含まれる共通接頭辞をすべて取得する
	#matcherには未登録語を処理するためのユーザ関数を指定する
	public function getall($text, $matcher) {
		if (!is_callable($matcher)) {
			trigger_error("ユーザ関数を実行できません", E_USER_WARNING);
			return;
		}
		$result = array();
		$matches = array();
		$enc = "UTF-8";
		while (mb_strlen($text, $enc) > 0) {
			$ret = call_user_func($matcher, $text);
			$len = mb_strlen($ret, $enc);
			$result = $this->commonprefix($text);
			if (count($result) > 0) {
				$end = end($result);
				if ($len > mb_strlen($end, $enc)) {
					$matches[] = array(mb_substr($text, 0, 1, $enc), $ret);
				} else {
					$matches[] = $result;
				}
			} elseif ($len > 1) {
				$matches[] = array(mb_substr($text, 0, 1, $enc), $ret);
			} else {
				$matches[] = array(mb_substr($text, 0, 1, $enc));
			}
			$text = mb_substr($text, 1, null, $enc);
		}
		return $matches;
	}

	public function save($fn) {
		$s = serialize($this);
  		$fp = fopen($fn, "w");
  		fwrite($fp, $s);
  		fclose($fp);
	}

	public function load($fn) {
		if (!file_exists($fn)) {return FALSE;}
		$content = file_get_contents($fn);
		$dummy = unserialize($content);
		$this->goto = $dummy->goto;
		$this->out = $dummy->out;
		$this->unum = $dummy->unum;
		return TRUE;
	}

	public function generate($fn) {
		if (!file_exists($fn)) {return FALSE;}
		$fp = fopen($fn,"r");
		while (($line = fgetss($fp))) {
			$line = preg_replace('/\A[\x00\s]++|[\x00\s]++\z/u', '', $line);
			if($line == "") {continue;}
			list($key, $value) = explode("\t", $line);
			$this->add($key, $value);
		}
		fclose($fp);
		return TRUE;
	}

	public function showMember() {
		print_r($this->goto);
		print_r($this->out);
	}
}


?>